#include <windows.h>
#include <stdio.h>
#include "args.h"

#ifdef _DEBUG
#define KdPrint(x)  {printf x;}
#else
#define KdPrint(x)  {;}
#endif

void print_help() {
    printf("Usage:\n"
           "  bbcopy.exe [-<switches>] <source file> <target file>\n"
           "Switches:\n"
           "  -b <blocksize>      sets physical block size for media. 2048 is default.\n"
           "  -o[b|k|m] <offset>  offset in bytes (or optionally in blocks, KBytes or MBytes)\n"
           "                      to start recovery from\n"
           "  -c[b|k|m] <count>   number of bytes (or optionally in blocks, KBytes or MBytes)\n"
           "                      to recover\n"
           "  -r <retry>          number of retry attempts. 3 is default\n"
           "  -l <logfile>        log unreliable and bad blocks to <logfile>\n"
           "  -s                  fill unreadable parts with \'** BAD BLOCK ***\'\n"
           "  -S <string>         fill unreadable parts with <string>\n"
           "                      Do not use fill patterns if you restore from multiple copies\n"
           "  -?                  display this help\n"
           "Note:\n"
           "  1 KByte = 1024 bytes\n"
           "  1 MByte = 1024 KBytes = 1024*1024 bytes\n"
           "  byte offset or byte count is specified, it will be rounded to blocksize\n"
           "Examples:\n"
           "  bbcopy.exe -o 5000000 X:\\Video.dat E:\\Temp\\Restored_Film.dat\n"
           "    will copy data starting from byte offset 5000000 to end of file\n"
           "  bbcopy.exe -om 1300 -cm 1000 X:\\Video.dat E:\\Temp\\Restored_Film.dat\n"
           "    will copy not more than 1000Mb of data starting from byte offset 1300Mb\n"
          );
    exit(0);
} // end print_help()

__int64
get_file_size(
    HANDLE h
    )
{
    LONG hsz = 0;
    LONG lsz;
    __int64 sz;

    lsz = SetFilePointer(h, 0, &hsz, FILE_END);
    if(lsz != -1 && hsz != -1) {
        ((PULONG)(&sz))[1] = hsz;
        ((PULONG)(&sz))[0] = lsz;
        return sz;
    }
    return -1;
} // end get_file_size()

__int64
set_file_pointer(
    HANDLE h,
    __int64 sz
    )
{
    ULONG hsz = ((PULONG)(&sz))[1];
    ULONG lsz = ((PULONG)(&sz))[0];

    lsz = SetFilePointer(h, lsz, (PLONG)&hsz, FILE_BEGIN);
    if(lsz != -1 && hsz != -1) {
        ((PULONG)(&sz))[1] = hsz;
        ((PULONG)(&sz))[0] = lsz;
        return sz;
    }
    return 0;
} // end set_file_pointer()

int
set_file_size(
    HANDLE h,
    __int64 sz
    )
{
    set_file_pointer(h, sz);
    return SetEndOfFile(h);
} // end set_file_size()

void
fill_buffer(
    char* buff,
    ULONG bs,
    PWCHAR fill_stringW,
    PCHAR fill_stringA
    )
{
    if(!fill_stringA && !fill_stringW) {
        memset(buff, 0, bs);
    } else {
        int tail = bs;
        int d;
        char* p_buff = buff;

        while(tail > 0) {
            if(GetVersion() & 0x80000000) {
                d = _snprintf(p_buff, tail, "%s", fill_stringA);
            } else {
                d = _snprintf(p_buff, tail, "%S", fill_stringW);
            }
            tail -= d;
            p_buff += d;
        }
    }
} // end fill_buffer()

void main (void) {
    int i, j;
    WCHAR a;
    ULONG bs = 2048;
    ULONG retry = 3;
    ULONG r;

    PWCHAR in_nameW = NULL;
    PWCHAR out_nameW = NULL;
    PWCHAR log_nameW = NULL;
    PWCHAR fill_stringW = NULL;

    PCHAR in_nameA = NULL;
    PCHAR out_nameA = NULL;
    PCHAR log_nameA = NULL;
    PCHAR fill_stringA = NULL;
    
    int break_sw;
    int block_offs = 0;
    int block_counter = 0;
    int new_target = 0;
    __int64 offs = 0;
    __int64 counter = 0;
    __int64 f_size;
    __int64 f_ptr;
    __int64 bads = 0;
    __int64 susp = 0;
    __int64 copy_count = 0;
    char* buff;
    ULONG read_bytes;
    BOOLEAN ok;
    __int64 p;

    WCHAR* CmdLineW;
    PWCHAR* argvW;
    CHAR*  CmdLineA;
    PCHAR* argvA;
    int argc;

    CHAR LogBuffer[2048];

    HANDLE in_h;
    HANDLE out_h;
    HANDLE log_h = NULL;

    printf("Scratched File Copy utility v1.7 (c) by Alter, 2004-2007\n"
           "Home site http://www.alter.org.ua\n");

    if(GetVersion() & 0x80000000) {
        printf("Warning: running under Windows 9x family\n"
               "  You may experience 2Gb or 4Gb limitation.\n");
//        exit(1);
        CmdLineA = GetCommandLineA();
        argvA = CommandLineToArgvA(CmdLineA, &argc);
#define argv     argvA
#define in_name  in_nameA
#define out_name out_nameA
#define log_name log_nameA
#define fill_string  fill_stringA
#define xcslen   strlen
#define xscanf   sscanf
#define xprintf  printf
#define CreateFileX   CreateFileA
#define XT(a)    a

        #include "parse_args.cpp.h"

    } else {
        CmdLineW = GetCommandLineW();
        argvW = CommandLineToArgvW(CmdLineW, &argc);
#pragma warning(push)               
#pragma warning(disable:4005)
//#pragma warning(disable:6311)
#define argv     argvW
#define in_name  in_nameW
#define out_name out_nameW
#define log_name log_nameW
#define fill_string  fill_stringW
#define xcslen   wcslen
#define xscanf   swscanf
#define xprintf  wprintf
#define CreateFileX   CreateFileW
#define XT(a)    L##a
#pragma warning(pop)

        #include "parse_args.cpp.h"
    
    }

    buff = (char*)GlobalAlloc(GMEM_DISCARDABLE, bs);
    if(!buff) {
        printf("Can't alloc %d bytes for temporary buffer\n"
               "exiting...\n", bs);
        exit (-2);
    }

    f_size = get_file_size(in_h);
    if(!f_size) {
        if(GetVersion() & 0x80000000) {
            printf("Empty source file %s\n"
                   "exiting...\n", out_nameA);
        } else {
            wprintf(L"Empty source file %s\n"
                   L"exiting...\n", out_name);
        }
        exit (-4);
    }
    if(f_size < 0) {
        if(GetVersion() & 0x80000000) {
            printf("Can't source file size: %s\n"
                   "exiting...\n", out_nameA);
        } else {
            wprintf(L"Can't source file size: %s\n"
                   L"exiting...\n", out_name);
        }
        exit (-4);
    }

    if(!set_file_size(out_h, f_size)) {
        if(GetVersion() & 0x80000000) {
            printf("Can't set size of target file %s to %I64d bytes\n"
                   "exiting...\n", out_nameA, f_size);
        } else {
            wprintf(L"Can't set size of target file %s to %I64d bytes\n"
                   L"exiting...\n", out_name, f_size);
        }
        exit (-5);
    }

    // align offs and counter
    offs = offs & ~((unsigned __int64)(bs-1));
    counter = (counter + bs - 1) & ~((unsigned __int64)(bs-1));

    f_ptr = offs;
    if(block_offs) {
        f_ptr *= bs;
    }
    if(block_counter) {
        counter *= bs;
    }

    if(GetVersion() & 0x80000000) {
        printf("%s -> %s\n%I64d bytes\n",
               in_nameA, out_nameA, f_size);
    } else {
        wprintf(L"%s -> %s\n%I64d bytes\n",
               in_name, out_name, f_size);
    }
    printf("Starting from offset %I64d bytes\n",
           offs);
    if(counter) {
        printf("Recovery range %I64d bytes\n",
               counter);
    }

    p = f_size/50;
    if(!p) p = 1;

    while(f_ptr <= f_size-bs) {
        if(counter && copy_count >= counter) {
            printf("Copied %I64d bytes, %I64d unreliable, %I64d in bad blocks\n", copy_count, susp, bads);
            exit(0);
        }
        if((f_ptr % p) < bs)
            printf(".");

        #include "copy_fragment.h"
    }

    printf("\n");
    if(f_size < f_ptr ||
        f_size - f_ptr > bs) {
        printf("Can't copy file tail (%I64d bytes)\n", f_size - f_ptr);
        exit(-7);
    }
    bs = (ULONG)(min(bs, f_size - f_ptr));
    if(!bs || (counter && copy_count >= counter)) {
        printf("Copied %I64d bytes, %I64d unreliable, %I64d in bad blocks\n", copy_count, susp, bads);
        exit(0);
    }

/*
    fill_buffer(buff, bs, fill_stringW, fill_stringA);
    set_file_pointer(in_h, f_ptr);
    set_file_pointer(out_h, f_ptr);
    ReadFile(in_h, buff, bs, &read_bytes, NULL);
    if(read_bytes != bs) {
        printf("Can't read block at offset %I64d\n", f_ptr);
        bads += bs;
    }
    WriteFile(out_h, buff, bs, &read_bytes, NULL);
    if(read_bytes != bs) {
        printf("Can't write block at offset %I64d\n", f_ptr);
    }
    copy_count += bs;
*/

    #define set_filler set_filler2
    #include "copy_fragment.h"

    printf("Copied %I64d bytes, %I64d unreliable, %I64d in bad blocks\n", copy_count, susp, bads);
    exit(0);
}
